/***
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    Project: www.simpledbm.org
 *    Author : Dibyendu Majumdar
 *    Email  : dibyendu@mazumdar.demon.co.uk
 */
/*
 * Created on: 05-Dec-2005
 * Author: Dibyendu Majumdar
 */
package org.simpledbm.samples.transaction;

import java.nio.ByteBuffer;

import org.simpledbm.common.registry.ObjectFactory;
import org.simpledbm.common.util.ByteString;
import org.simpledbm.common.util.TypeSize;
import org.simpledbm.rss.bm.BufMgr;
import org.simpledbm.rss.bm.BufMgrException;
import org.simpledbm.rss.bm.BufferAccessBlock;
import org.simpledbm.rss.io.StorageContainer;
import org.simpledbm.rss.io.StorageContainerFactory;
import org.simpledbm.rss.io.StorageException;
import org.simpledbm.rss.io.StorageManager;
import org.simpledbm.rss.locking.LockDuration;
import org.simpledbm.rss.locking.LockMode;
import org.simpledbm.rss.log.LogMgr;
import org.simpledbm.rss.log.Lsn;
import org.simpledbm.rss.pm.Page;
import org.simpledbm.rss.pm.PageException;
import org.simpledbm.rss.pm.PageFactory;
import org.simpledbm.rss.pm.PageId;
import org.simpledbm.rss.sm.SpaceMgrException;
import org.simpledbm.rss.tm.BaseLoggable;
import org.simpledbm.rss.tm.BaseTransactionalModule;
import org.simpledbm.rss.tm.Compensation;
import org.simpledbm.rss.tm.ContainerDeleteOperation;
import org.simpledbm.rss.tm.Loggable;
import org.simpledbm.rss.tm.LoggableFactory;
import org.simpledbm.rss.tm.NonTransactionRelatedOperation;
import org.simpledbm.rss.tm.PageFormatOperation;
import org.simpledbm.rss.tm.PostCommitAction;
import org.simpledbm.rss.tm.Redoable;
import org.simpledbm.rss.tm.Transaction;
import org.simpledbm.rss.tm.TransactionalModuleRegistry;
import org.simpledbm.rss.tm.TrxException;
import org.simpledbm.rss.tm.TrxMgr;
import org.simpledbm.rss.tm.Undoable;

/**
 * Demonstrates how to build a Transactional Module. The example here is based upon
 * the description of a One-Bit Resource Manager in section 10.3.7.2 of the classic book
 * Transaction Processing: Concepts and Techniques by Jim Gray and Andreas Reuter.
 * <p>
 * The One Bit Manager manages setting and unsetting of bits within a single page
 * of a container. Its operations are transactional, and demonstrate several concepts
 * related to transaction processing, such as locking, write ahead logging, commit and
 * rollback processing, etc.
 * <p>
 * The Bit Manager supports following transactional actions:
 * <ol>
 * <li>Create a new container. The container will contain a single page.</li>
 * <li>Set a bit. The bit will be exclusively locked as a result.</li>
 * <li>Read a bit. The bit will be locked in shared mode during the read.</li>
 * <li>Read a bit in update mode. This causes the bit to be locked in Update mode.</li>
 * <li>Drop a container.</li>
 * </ol>
 * <p>
 * All of above operations are transactional and therefore show ACID properties.
 *
 * @author Dibyendu Majumdar
 * @since 05-Dec-2005
 */
public class OneBitMgr extends BaseTransactionalModule {

    /**
     * Each transactional module must have its own unique ID. 
     */
    private static final short MODULE_ID = 250;

    static final short TYPE_BITMGR_PAGE = 25000;
    static final short TYPE_BITMGR_LOGFORMATPAGE = 25001;
    static final short TYPE_BITMGR_LOGCHANGEBIT = 25002;
    static final short TYPE_BITMGR_LOGUNDOCHANGEBIT = 25003;
    static final short TYPE_BITMGR_LOGCREATECONTAINER = 25004;
    static final short TYPE_BITMGR_LOGOPENCONTAINER = 25005;
    static final short TYPE_BITMGR_LOGUNDOCREATECONTAINER = 25006;
    static final short TYPE_BITMGR_LOGDROPCONTAINER = 25007;
    
    static final int PAGE_NUMBER = 0;
    
    BufMgr bufmgr;
    
    LogMgr logmgr;
    
    LoggableFactory loggableFactory;

    StorageContainerFactory storageFactory;
    
    StorageManager storageManager;
    
    TrxMgr trxmgr;

    PageFactory pageFactory;
    
    ObjectFactory objectFactory;

    public OneBitMgr(StorageContainerFactory storageFactory, StorageManager storageManager, PageFactory pageFactory, BufMgr bufmgr, LoggableFactory loggableFactory, TrxMgr trxmgr, ObjectFactory objectFactory, TransactionalModuleRegistry moduleRegistry) {
        this.storageFactory = storageFactory;
        this.storageManager = storageManager;
        this.pageFactory = pageFactory;
        this.bufmgr = bufmgr;
        this.loggableFactory = loggableFactory;
        this.objectFactory = objectFactory;
        this.trxmgr = trxmgr;

        /* 
         * A Transactional Module must register itself with the TransactionalModuleRegistry.
         * This enables the Transaction Manager to locate the module.
         */
        moduleRegistry.registerModule(MODULE_ID, this);
        
        /*
         * Typically, a transactional module will have its own set of object types. These may
         * be log records, disk page types, etc. It is good practice to register all the
         * module specific objects in the constructor.
         */
        objectFactory.register(TYPE_BITMGR_PAGE, BitMgrPage.class.getName());
        objectFactory.register(TYPE_BITMGR_LOGCREATECONTAINER, CreateContainer.class.getName());
        objectFactory.register(TYPE_BITMGR_LOGFORMATPAGE, BitMgrLogFormatPage.class.getName());
        objectFactory.register(TYPE_BITMGR_LOGCHANGEBIT, BitMgrLogChangeBit.class.getName());
        objectFactory.register(TYPE_BITMGR_LOGUNDOCHANGEBIT, BitMgrUndoChangeBit.class.getName());
        objectFactory.register(TYPE_BITMGR_LOGUNDOCREATECONTAINER, UndoCreateContainer.class.getName());
        objectFactory.register(TYPE_BITMGR_LOGDROPCONTAINER, DropContainer.class.getName());
        objectFactory.register(TYPE_BITMGR_LOGOPENCONTAINER, OpenContainer.class.getName());
    }
    
    /**
     * Demonstrates page oriented undo, where the undo operation applies to the
     * same page that was affected during forward processing.
     * 
     * @see org.simpledbm.rss.tm.TransactionalModule#generateCompensation(org.simpledbm.rss.tm.Undoable)
     */
    @Override
    public final Compensation generateCompensation(Undoable undoable) {
        if (undoable instanceof BitMgrLogChangeBit) {
            BitMgrUndoChangeBit clr = (BitMgrUndoChangeBit) loggableFactory.getInstance(MODULE_ID, TYPE_BITMGR_LOGUNDOCHANGEBIT);
            BitMgrLogChangeBit logrec = (BitMgrLogChangeBit) undoable;
            clr.index = logrec.index;
            clr.newValue = logrec.oldValue;
            return clr;
        }
        else if (undoable instanceof CreateContainer) {
            /*
             * Strictly speaking container create operation is not page oriented. However,
             * since the Transaction Manager can only track changes at a page level, this
             * type of operation needs to be logged against a page. One choice would be to log
             * it against a data dictionary page. Unfortunately, that would create a dependency between
             * any system that wants to create a container and the data dictionary. The approach in
             * SimpleDBM is to log container operations such as this against a "virtual" page
             * located at PageId(0,0). This page belongs to a container that is created when
             * a SimpleDBM instance is created - hence it is guaranteed to be present.
             */
            UndoCreateContainer clr = (UndoCreateContainer) loggableFactory.getInstance(MODULE_ID, TYPE_BITMGR_LOGUNDOCREATECONTAINER);
            CreateContainer logrec = (CreateContainer) undoable;
            clr.containerId = logrec.containerId;
            clr.name = logrec.name;
            return clr;
        }
        return null;
    }

   /**
    * Demonstrates redo of non-page related operations. Note that the Transaction 
    * Manager will ignore exceptions thrown by this method. 
    * 
    * @see org.simpledbm.rss.tm.TransactionalModule#redo(org.simpledbm.rss.tm.Loggable)
    */
    @Override
    public final void redo(Loggable loggable) throws StorageException, TrxException {
        if (loggable instanceof OpenContainer) {
            OpenContainer logrec = (OpenContainer) loggable;
            StorageContainer sc = storageManager.getInstance(logrec.getContainerId()); 
            if (sc == null) {
                sc = storageFactory.open(logrec.getName());
                storageManager.register(logrec.getContainerId(), sc);
            }
        }
        else if (loggable instanceof DropContainer) {
            DropContainer logrec = (DropContainer) loggable;
            /* 
             * Before dropping the container we need to ensure that the
             * Buffer Manager does not hold references to any pages that are
             * part of this container.
             */ 
            bufmgr.invalidateContainer(logrec.getContainerId());
            storageManager.remove(logrec.getContainerId());
            storageFactory.delete(logrec.getName());
            /*
             * Since this is a PostCommitAction, the log merely records that this
             * operation has been performed. 
             */
            trxmgr.logNonTransactionRelatedOperation(logrec);
        }
    }
    
    /**
     * Demonstrates page oriented redo operations.
     *  
     * @see org.simpledbm.rss.tm.TransactionalModule#redo(org.simpledbm.rss.pm.Page, org.simpledbm.rss.tm.Redoable)
     */
    @Override
    public void redo(Page page, Redoable loggable) throws StorageException, PageException {
        if (loggable instanceof CreateContainer) {
            /*
             * Container creation is logged against a "virtual" page located at PageId(0,0).
             */
            CreateContainer logrec = (CreateContainer) loggable;
            StorageContainer sc = storageFactory.create(logrec.getName());
            storageManager.register(logrec.getContainerId(), sc);
        }
        else if (loggable instanceof UndoCreateContainer) {
            UndoCreateContainer logrec = (UndoCreateContainer) loggable;
            /* 
             * Before dropping the container we need to ensure that the
             * Buffer Manager does not hold references to any pages that are
             * part of this container.
             */ 
            bufmgr.invalidateContainer(logrec.getContainerId());
            storageManager.remove(logrec.getContainerId());
            storageFactory.delete(logrec.getName());
        }
        else if (loggable instanceof BitMgrLogChangeBit) {
            /*
             * Redo an update to a bit.
             */
            BitMgrLogChangeBit logrec = (BitMgrLogChangeBit) loggable;
            System.out.println("DO: Setting bit[" + logrec.index + "] to " + logrec.newValue);
            BitMgrPage bpage = (BitMgrPage) page;
            bpage.setBit(logrec.index, (byte) logrec.newValue);
        }
        else if (loggable instanceof BitMgrLogFormatPage) {
            /*
             * Page format operations are externalized immediately. This
             * is optional, but recommended as any storage issues can be 
             * discovered before the page is actually used.
             */
            page.setPageLsn(loggable.getLsn());
            pageFactory.store(page);
        }
        else if (loggable instanceof BitMgrUndoChangeBit) {
            /*
             * Undo an update to a bit.
             */
            BitMgrUndoChangeBit logrec = (BitMgrUndoChangeBit) loggable;
            System.out.println("UNDO: Setting bit[" + logrec.index + "] to " + logrec.newValue);
            BitMgrPage bpage = (BitMgrPage) page;
            bpage.setBit(logrec.index, (byte) logrec.newValue);
        }
    }

    
    
    /**
     * 
     * @param trx
     * @param index
     * @param newValue
     * @throws BufMgrException
     * @throws TrxException
     * @throws StorageException
     * @throws PageException
     */
    public void changeBit(int containerId, Transaction trx, int index, int newValue) throws BufMgrException, TrxException, StorageException, PageException {
        trx.acquireLock(new ObjectLock(ObjectLock.CONTAINER, containerId), LockMode.SHARED, LockDuration.MANUAL_DURATION);
        trx.acquireLock(new ObjectLock(ObjectLock.BIT, index), LockMode.EXCLUSIVE, LockDuration.MANUAL_DURATION);
        BitMgrLogChangeBit logrec = (BitMgrLogChangeBit) loggableFactory.getInstance(MODULE_ID, TYPE_BITMGR_LOGCHANGEBIT);
        BufferAccessBlock bab = bufmgr.fixExclusive(new PageId(containerId, PAGE_NUMBER), false, TYPE_BITMGR_PAGE, 0);
        try {
            BitMgrPage page = (BitMgrPage) bab.getPage();
            
            logrec.index = index;
            logrec.newValue = newValue;
            logrec.oldValue = page.getBit(index);
            
            Lsn lsn = trx.logInsert(page, logrec);
            redo(page, logrec);
            bab.setDirty(lsn);
        }
        finally {
            bab.unfix();
        }
    }
    
    public int getBit(int containerId, Transaction trx, int index) throws BufMgrException, TrxException {
        trx.acquireLock(new ObjectLock(ObjectLock.CONTAINER, containerId), LockMode.SHARED, LockDuration.MANUAL_DURATION);
        trx.acquireLock(new ObjectLock(ObjectLock.BIT, index), LockMode.SHARED, LockDuration.MANUAL_DURATION);
        BufferAccessBlock bab = bufmgr.fixShared(new PageId(containerId, PAGE_NUMBER), 0);
        int value = -1;
        try {
            BitMgrPage page = (BitMgrPage) bab.getPage();
            value = page.getBit(index);
        }
        finally {
            bab.unfix();
        }
        trx.releaseLock(new ObjectLock(ObjectLock.BIT, index));
        trx.releaseLock(new ObjectLock(ObjectLock.CONTAINER, containerId));
        return value;
    }
    
    public int getBitForUpdate(int containerId, Transaction trx, int index) throws BufMgrException, TrxException {
        trx.acquireLock(new ObjectLock(ObjectLock.CONTAINER, containerId), LockMode.SHARED, LockDuration.MANUAL_DURATION);
        trx.acquireLock(new ObjectLock(ObjectLock.BIT, index), LockMode.UPDATE, LockDuration.MANUAL_DURATION);
        BufferAccessBlock bab = bufmgr.fixShared(new PageId(containerId, PAGE_NUMBER), 0);
        int value = -1;
        try {
            BitMgrPage page = (BitMgrPage) bab.getPage();
            value = page.getBit(index);
        }
        finally {
            bab.unfix();
        }
        return value;
    }
    
    public void create(String name, int containerId) throws StorageException, BufMgrException, TrxException, PageException {
        
        final int pageNumber = 0;
        Transaction trx = trxmgr.begin();
        boolean success = false;
        try {
            CreateContainer logcreate = (CreateContainer) loggableFactory.getInstance(MODULE_ID, TYPE_BITMGR_LOGCREATECONTAINER);
            OpenContainer logopen = (OpenContainer) loggableFactory.getInstance(MODULE_ID, TYPE_BITMGR_LOGOPENCONTAINER);
            BitMgrLogFormatPage formatpage = (BitMgrLogFormatPage) loggableFactory.getInstance(MODULE_ID, TYPE_BITMGR_LOGFORMATPAGE);
            
            trx.acquireLock(new ObjectLock(ObjectLock.CONTAINER, containerId), LockMode.EXCLUSIVE, LockDuration.MANUAL_DURATION);
            /* check if the container already exists */
            StorageContainer sc = storageManager.getInstance(containerId);
            if (sc != null) {
                throw new StorageException(); // TODO change exception type
            }
            trx.startNestedTopAction();
            
            BufferAccessBlock bab1 = bufmgr.fixExclusive(new PageId(0, 0), false, -1, 0);
            try {
                Page page1 = bab1.getPage();
                logcreate.setContainerId(containerId);
                logcreate.setName(name);
                Lsn mylsn = trx.logInsert(page1, logcreate);
                redo(page1, logcreate);
                bab1.setDirty(mylsn);

                logopen.setContainerId(containerId);
                logopen.setName(name);
                trxmgr.logNonTransactionRelatedOperation(logopen);
                
                PageId pageId = new PageId(containerId, pageNumber);
                formatpage.setPageId(TYPE_BITMGR_PAGE, pageId);
                BufferAccessBlock bab = bufmgr.fixExclusive(pageId, true, TYPE_BITMGR_PAGE, 0);
                try {
                    Page page = bab.getPage();
                    mylsn = trx.logInsert(page, formatpage);
                    redo(page, formatpage);
                    bab.setDirty(mylsn);
                }               
                finally {
                    bab.unfix();
                }
            }
            finally {
                bab1.unfix();
            }
            trx.completeNestedTopAction();
            success = true;
        }
        catch (StorageException e) {
            e.printStackTrace();
            throw e;
        }
        catch (BufMgrException e) {
            e.printStackTrace();
            throw e;
        }
        catch (TrxException e) {
            e.printStackTrace();
            throw e;
        }
        catch (PageException e) {
            e.printStackTrace();
            throw e;
        }
        finally {
            if (success) {
                trx.commit();
            }
            else {
                trx.abort();
            }
        }
    }

    /**
     * Drops the container specified. 
     */
    public void dropContainer(int containerId) throws SpaceMgrException,
            TrxException {

        Transaction trx = trxmgr.begin();
        boolean success = false;
        try {
            trx.acquireLock(new ObjectLock(ObjectLock.CONTAINER, containerId),
                    LockMode.EXCLUSIVE, LockDuration.MANUAL_DURATION);
            DropContainer logrec = (DropContainer) loggableFactory.getInstance(
                    MODULE_ID, TYPE_BITMGR_LOGDROPCONTAINER);
            StorageContainer sc;
            try {
                sc = storageManager.getInstance(containerId);
            } catch (StorageException e) {
                throw new SpaceMgrException("Container " + containerId
                        + " cannot be accessed", e);
            }
            if (sc == null) {
                throw new SpaceMgrException("Container " + containerId
                        + " does not exist");
            } else {
                logrec.setContainerId(containerId);
                logrec.setName(sc.getName());
                trx.schedulePostCommitAction(logrec);
            }
            success = true;
        } finally {
            if (success) {
                trx.commit();
            } else {
                trx.abort();
            }
        }
    }
    
    /**
     * Represents a lock object. 
     *
     * @author Dibyendu Majumdar
     * @since 05-Dec-2005
     */
    public static class ObjectLock {
        
        static final int CONTAINER = 1;
        static final int BIT = 2;
        
        final int type;
        final int value;
        
        public ObjectLock(int type, int value) {
            this.type = type;
            this.value = value;
        }

        @Override
        public boolean equals(Object obj) {
            if (obj instanceof ObjectLock) {
                ObjectLock lock = (ObjectLock) obj;
                return lock.type == type && lock.value == value;
            }
            return false;
        }

        @Override
        public int hashCode() {
            return type ^ value;
        }

        @Override
        public String toString() {
            return "ObjectLock(type=" + type + ", value = " + value + ")";
        }
    }

    /**
     * Base class for container operations.
     */
    static abstract class ContainerOperation extends BaseLoggable {

        int containerId;
        ByteString name;
        
        @Override
        public final void init() {
            containerId = -1;
            name = new ByteString("");
        }
        
        public final int getContainerId() {
            return containerId;
        }

        public final void setContainerId(int containerId) {
            this.containerId = containerId;
        }

        public final String getName() {
            return name.toString();
        }

        public final void setName(String name) {
            this.name = new ByteString(name);
        }

        @Override
        public int getStoredLength() {
            return super.getStoredLength() + TypeSize.INTEGER + name.getStoredLength();
        }

        @Override
        public void retrieve(ByteBuffer bb) {
            super.retrieve(bb);
            containerId = bb.getInt();
            name = new ByteString();
            name.retrieve(bb);
        }

        @Override
        public void store(ByteBuffer bb) {
            super.store(bb);
            bb.putInt(containerId);
            name.store(bb);
        }
        
        @Override
        public String toString() {
            return "containerId=" + containerId + ", name=" + name.toString() + "," + super.toString();
        }
    }
    
    /**
     * Log for recording the creation of a container. This is an Undoable operation, 
     * therefore if the transaction that created the container aborts, the newly created
     * container will be dropped.
     */
    public static final class CreateContainer extends ContainerOperation implements Undoable {
        @Override
        public String toString() {
            return "CreateContainer(" + super.toString() + ")";
        }
    }
    
    /**
     * Log for undoing a container create operation. After this is executed, all pages in the 
     * Buffer Manager for this container should be marked invalid.
     */
    public static final class UndoCreateContainer extends ContainerOperation implements Compensation, ContainerDeleteOperation {        
        @Override
        public String toString() {
            return "UndoCreateContainer(" + super.toString() + ")";
        }
    }   
    
    /**
     * Logs the fact that a new container has been created and must be opened. 
     * This ensures that that the container is reopened during the redo phase.
     */
    public static final class OpenContainer extends ContainerOperation implements NonTransactionRelatedOperation {
        @Override
        public String toString() {
            return "OpenContainer(" + super.toString() + ")";
        }
    }   
    
    /**
     * Log record for container delete operation.  
     */
    public static final class DropContainer extends ContainerOperation implements PostCommitAction, ContainerDeleteOperation {
        
        /**
         * Unique transaction specific action id to track the status of the operation.
         */
        int actionId;
        
        public final int getActionId() {
            return actionId;
        }

        public final void setActionId(int actionId) {
            this.actionId = actionId;
        }

        @Override
        public void retrieve(ByteBuffer bb) {
            super.retrieve(bb);
            actionId = bb.getInt();
        }

        @Override
        public void store(ByteBuffer bb) {
            super.store(bb);
            bb.putInt(actionId);
        }

        @Override
        public int getStoredLength() {

            int retValue;
            
            retValue = super.getStoredLength();
            retValue += TypeSize.INTEGER;
            return retValue;
        }

        @Override
        public String toString() {
            return "DropContainer(actionId=" + actionId + ", " + super.toString() + ")";
        }
    
    }
    
    public static class BitMgrLogFormatPage extends BaseLoggable implements Redoable, PageFormatOperation {

        @Override
        public void init() {
        }
        
    }
    
    public static class ChangeBitRedo extends BaseLoggable {

        int index;
        int newValue;
        
        @Override
        public void init() {
        }

        public final int getIndex() {
            return index;
        }

        public final void setIndex(int index) {
            this.index = index;
        }

        public final int getNewValue() {
            return newValue;
        }

        public final void setNewValue(int newValue) {
            this.newValue = newValue;
        }

        @Override
        public int getStoredLength() {
            return super.getStoredLength() + TypeSize.INTEGER * 2;
        }

        @Override
        public void retrieve(ByteBuffer bb) {
            super.retrieve(bb);
            index = bb.getInt();
            newValue = bb.getInt();
        }

        @Override
        public void store(ByteBuffer bb) {
            super.store(bb);
            bb.putInt(index);
            bb.putInt(newValue);
        }
    }

    public static class ChangeBitRedoUndo extends ChangeBitRedo {

        int oldValue;
        
        @Override
        public void init() {
        }

        public final int getOldValue() {
            return oldValue;
        }

        public final void setOldValue(int oldValue) {
            this.oldValue = oldValue;
        }

        @Override
        public int getStoredLength() {
            return super.getStoredLength() + TypeSize.INTEGER;
        }

        @Override
        public void retrieve(ByteBuffer bb) {
            super.retrieve(bb);
            oldValue = bb.getInt();
        }

        @Override
        public void store(ByteBuffer bb) {
            super.store(bb);
            bb.putInt(oldValue);
        }
    }

    public static class BitMgrLogChangeBit extends ChangeBitRedoUndo implements Undoable {
    }

    public static class BitMgrUndoChangeBit extends ChangeBitRedo implements Compensation {
    }
    
    public static class BitMgrPage extends Page {

        byte[] bits;
        
        @Override
        public void init() {
            int n_bits = super.getStoredLength() - Page.SIZE;
            bits = new byte[n_bits];
        }

        @Override
        public void retrieve(ByteBuffer bb) {
            super.retrieve(bb);
            bb.get(bits);
        }

        @Override
        public void store(ByteBuffer bb) {
            super.store(bb);
            bb.put(bits);
        }
        
        public void setBit(int offset, byte value) {
            bits[offset] = value;
        }
        
        public int getBit(int offset) {
            return bits[offset];
        } 
    }
    
}
