/***
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    Project: www.simpledbm.org
 *    Author : Dibyendu Majumdar
 *    Email  : dibyendu@mazumdar.demon.co.uk
 */
package org.simpledbm.rss.bm;

import org.simpledbm.rss.log.Lsn;
import org.simpledbm.rss.pm.PageId;

/**
 * Defines the interface for the Buffer Manager module. 
 * 
 * @author Dibyendu Majumdar
 * @since 18-Aug-2005
 */
public interface BufMgr {

	/**
	 * Start the Buffer Manager instance.
	 */
	public void start() throws BufMgrException;
	
    /**
     * Shuts down the Buffer Manager instance.
     */
    public void shutdown();

    /**
     * Fixes and latches a page in buffer pool in shared mode.  
     * It is an error if the page does not already exist in persistent storage.
     * 
     * @param pageid
     *            The identity of the page that should be fixed.
     * @param hint A hint to indicate which end of the LRU chain the page should be
     *             inserted to.
     * @return A BufferAccessBlock containing a reference to the desired page.
     * @throws BufMgrException
     * @see BufferAccessBlock#unfix()
     */
    public BufferAccessBlock fixShared(PageId pageid, int hint) throws BufMgrException;
    
    /**
     * Fixes and latches a page in buffer pool in exclusive mode.  
     * <p>
     * A page that is latched exclusively may be downgraded to update mode
     * if desired by calling {@link BufferAccessBlock#downgradeExclusiveLatch()}.
     * <p>
     * Unless a request is being made to fix a new page, the page must already
     * exist in persistent storage.
     * Caller must ensure that request for a new page is not made for
     * a page already in the buffer pool.
     * 
     * @param pageid
     *            The identity of the page that should be fixed.
     * @param isNew
     *            If this is set to true, the page will not be read from disk.
     * @param pagetype
     *            Specifies the type of page to create; only used when isNew is
     *            set.
     * @param hint A hint to indicate which end of the LRU chain the page should be
     *             inserted to.
     * @return A BufferAccessBlock containing a reference to the desired page.
     * @throws BufMgrException
     * @see BufferAccessBlock#unfix()
     * @see BufferAccessBlock#downgradeExclusiveLatch()
     * @see BufferAccessBlock#setDirty(Lsn)
     */
    public BufferAccessBlock fixExclusive(PageId pageid, boolean isNew, int pagetype,
            int hint) throws BufMgrException;

    /**
     * Fixes and latches a page in buffer pool fir Update.  
     * A page that is latched in update mode can be upgraded to exclusive mode
     * by calling {@link BufferAccessBlock#upgradeUpdateLatch()}.
     * It is an error if the page does not already exist in persistent storage.
     * 
     * @param pageid
     *            The identity of the page that should be fixed.
     * @param hint A hint to indicate which end of the LRU chain the page should be
     *             inserted to.
     * @return A BufferAccessBlock containing a reference to the desired page.
     * @throws BufMgrException
     * @see BufferAccessBlock#unfix()
     * @see BufferAccessBlock#upgradeUpdateLatch()
     */
    public BufferAccessBlock fixForUpdate(PageId pageid, int hint) throws BufMgrException;
    
    /**
     * Obtains a list of dirty pages. Called by the Transaction Manager during
     * checkpoints. 
     * 
     * @return An array of pages that are dirty
     */
    public DirtyPageInfo[] getDirtyPages();

    /**
     * Allows the update of recoveryLsn after transaction manager has performed
     * recovery. This method is intended for use by the Transaction Manager
     * only. Since this method is called during restart recovery only, there is
     * no need to make it thread safe.
     * 
     * @param dirty_pages List of dirty pages as determined by Transaction Manager
     */
    public void updateRecoveryLsn(DirtyPageInfo[] dirty_pages);

    /**
     * Marks all pages of specified container as invalid. 
     * 
     * @param containerId ID of the container that is to be invalidated
     */
    void invalidateContainer(int containerId);    
}
