/***
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    Project: www.simpledbm.org
 *    Author : Dibyendu Majumdar
 *    Email  : dibyendu@mazumdar.demon.co.uk
 */
package org.simpledbm.rss.bm;

import org.simpledbm.rss.log.Lsn;
import org.simpledbm.rss.pm.Page;

/**
 * A BufferAccessBlock encapsulates access to a fixed (pinned) 
 * page. 
 *  
 * @author Dibyendu Majumdar
 * @since 18-Aug-2005
 */
public interface BufferAccessBlock {

	/**
	 * Marks the page as dirty. Only applicable if the
	 * page was obtained with exclusive access.
	 * @throws IllegalStateException If page is not held in exclusive latch mode
	 */
    public void setDirty(Lsn lsn);

    /**
     * Gets the page being accessed. It is assumed that the
     * caller will cast the page to appropriate type.
     * @return Page object
     */
    public Page getPage();
    
	/**
	 * Upgrades page latch from UPDATE to EXCLUSIVE.
	 * Latch must be held in UPDATE mode, else an Exception will
	 * be thrown.
	 * @throws IllegalStateException If latch is not held in update mode
	 */
    public void upgradeUpdateLatch();
    
	/**
	 * Downgrades page latch from EXCLUSIVE to UPDATE.
	 * Latch must be held in EXCLUSIVE mode, else an Exception will
	 * be thrown.
	 * @throws IllegalStateException If latch is not held exclusively
	 */
    public void downgradeExclusiveLatch();
    
    /**
     * Unfixes and releases latch on the page. It is important
     * that this is called for every fix. Failure to do so will cause
     * pages to get stuck in the buffer pool, eventually causing the
     * system to run of slots in the buffer pool.
     * 
     * @throws BufMgrException thrown if there was an unexpected error
     */
    public void unfix() throws BufMgrException;

    /**
     * Tests whether the page is currently latched in EXCLUSIVE mode.
     */
    public boolean isLatchedExclusively();
	
    /**
     * Tests whether the page is currently latched in UPDATE mode.
     */
	public boolean isLatchedForUpdate();
	
    /**
     * Tests whether the page is currently latched in SHARED mode.
     */
	public boolean isLatchedShared();

}
