package org.simpledbm.rss.main;

import java.util.Properties;

import org.simpledbm.common.latch.LatchFactory;
import org.simpledbm.common.latch.impl.LatchFactoryImpl;
import org.simpledbm.common.registry.ObjectFactory;
import org.simpledbm.common.registry.impl.ObjectFactoryImpl;
import org.simpledbm.common.util.logging.Logger;
import org.simpledbm.rss.bm.BufMgr;
import org.simpledbm.rss.bm.impl.BufMgrImpl;
import org.simpledbm.rss.bt.IndexManager;
import org.simpledbm.rss.bt.impl.BTreeMgrImpl;
import org.simpledbm.rss.io.StorageContainer;
import org.simpledbm.rss.io.StorageContainerFactory;
import org.simpledbm.rss.io.StorageManager;
import org.simpledbm.rss.io.impl.FileStorageContainerFactory;
import org.simpledbm.rss.io.impl.StorageManagerImpl;
import org.simpledbm.rss.locking.LockMgr;
import org.simpledbm.rss.locking.LockMgrFactory;
import org.simpledbm.rss.locking.impl.LockMgrFactoryImpl;
import org.simpledbm.rss.log.LogFactory;
import org.simpledbm.rss.log.LogMgr;
import org.simpledbm.rss.log.impl.LogFactoryImpl;
import org.simpledbm.rss.pm.Page;
import org.simpledbm.rss.pm.PageFactory;
import org.simpledbm.rss.pm.PageId;
import org.simpledbm.rss.pm.impl.PageFactoryImpl;
import org.simpledbm.rss.sm.SpaceMgr;
import org.simpledbm.rss.sm.impl.SpaceMgrImpl;
import org.simpledbm.rss.sp.SlottedPageMgr;
import org.simpledbm.rss.sp.impl.SlottedPageMgrImpl;
import org.simpledbm.rss.tm.LoggableFactory;
import org.simpledbm.rss.tm.TransactionalModuleRegistry;
import org.simpledbm.rss.tm.TrxMgr;
import org.simpledbm.rss.tm.impl.LoggableFactoryImpl;
import org.simpledbm.rss.tm.impl.TransactionalModuleRegistryImpl;
import org.simpledbm.rss.tm.impl.TrxMgrImpl;
import org.simpledbm.rss.tuple.TupleManager;
import org.simpledbm.rss.tuple.impl.TupleManagerImpl;

/**
 * A Server instance encapsulates all the modules that comprise the RSS. 
 * It ensures that all modules are initialized in the correct order and provides
 * a mechanism to start and stop the instance.
 * <p>
 * Note that the Server component acts very much like a custom IoC Container.  
 * 
 * @author Dibyendu Majumdar
 * @since 03-Apr-2006
 */
public class Server {

	private static String LOG_CLASS_NAME = Server.class.getName();
	private static Logger log = Logger.getLogger("org.simpledbm.rss.main");
	
	public static final String VIRTUAL_TABLE = "_internal/dual";
	public static final int VIRTUAL_TABLE_CONTAINER_ID = 0;
	
    final ObjectFactory objectRegistry;
    final StorageContainerFactory storageFactory;
    final StorageManager storageManager;
    final LatchFactory latchFactory;
    final PageFactory pageFactory;
    final SlottedPageMgr slottedPageManager;
    final LockMgr lockManager;
    final LogMgr logManager;
    final BufMgr bufferManager;
    final LoggableFactory loggableFactory;
    final TransactionalModuleRegistry moduleRegistry;
	final TrxMgr transactionManager;
    final SpaceMgr spaceManager;
    final IndexManager indexManager;
    final TupleManager tupleManager;
	
	/**
	 * Creates a new RSS Server instance. An RSS Server instance contains at least a 
	 * LOG instance, and a VIRTUAL TABLE called dual. 
	 * @see LogFactory#createLog(Properties)
	 * @see LogFactory
	 */
	public static void create(Properties props) throws Exception {
		Server server = new Server(props);
		final LogFactory logFactory = new LogFactoryImpl();
		logFactory.createLog(server.getStorageFactory(), props);
		// SimpleDBM components expect a virtual container to exist with
		// a container ID of 1. This container must have at least one page.
		StorageContainer sc = server.getStorageFactory().create(VIRTUAL_TABLE);
		server.getStorageManager().register(VIRTUAL_TABLE_CONTAINER_ID, sc);
		Page page = server.getPageFactory().getInstance(server.getPageFactory().getRawPageType(), new PageId(VIRTUAL_TABLE_CONTAINER_ID, 0));
		server.getPageFactory().store(page);
		// We start the server so that a checkpoint can be taken which will
		// ensure that the VIRTUAL_TABLE is automatically opened at system startup.
		server.start();
		server.shutdown();
	}

	/**
	 * Initializes a new RSS Server instance.
	 * <p>
	 * FIXME: Change exception specification.
	 * 
	 * @inheritDoc
	 * @see #start()
	 * @see #shutdown()
	 * @param props Properties that define various parameters for the system
	 * @throws Exception Thrown if there was a problem during initialization.
	 */
	public Server(Properties props) throws Exception {

		String logProperties = props.getProperty("logging.properties.file", "logging.properties");
		Logger.configure(logProperties);
		
		final LogFactory logFactory = new LogFactoryImpl();
		final LockMgrFactory lockMgrFactory = new LockMgrFactoryImpl();
		
		objectRegistry = new ObjectFactoryImpl();
		storageFactory = new FileStorageContainerFactory(props);
		storageManager = new StorageManagerImpl();
		latchFactory = new LatchFactoryImpl();
		pageFactory = new PageFactoryImpl(objectRegistry, storageManager, latchFactory);
		slottedPageManager = new SlottedPageMgrImpl(objectRegistry);
		loggableFactory = new LoggableFactoryImpl(objectRegistry);
		moduleRegistry = new TransactionalModuleRegistryImpl();
		lockManager = lockMgrFactory.create(props);
		logManager = logFactory.getLog(storageFactory, props);
		bufferManager = new BufMgrImpl(logManager, pageFactory, props);
		transactionManager = new TrxMgrImpl(logManager, storageFactory, storageManager, bufferManager, lockManager, loggableFactory, latchFactory, objectRegistry, moduleRegistry);
		spaceManager = new SpaceMgrImpl(objectRegistry, pageFactory, logManager, bufferManager, storageManager, storageFactory, loggableFactory, transactionManager, moduleRegistry);
		indexManager = new BTreeMgrImpl(objectRegistry, loggableFactory, spaceManager, bufferManager, slottedPageManager, moduleRegistry);
        tupleManager = new TupleManagerImpl(objectRegistry, loggableFactory, spaceManager, bufferManager, slottedPageManager, moduleRegistry, pageFactory);
	}

	/**
	 * Starts the Server instance. This results in following actions:
	 * <ol>
	 * <li>The Log instance is opened. This starts the background threads that manage log writes.</li>
	 * <li>The Buffer Manager instance is started. This starts up the Buffer Writer thread.</li>
	 * <li>The Transaction Manager is started. This initiates restart recovery, and also starts the
	 *     Checkpoint thread.</li>
	 * </ol>
	 * <p>
	 * FIXME: Exception signature needs to be fixed.
	 * 
	 * @see LogMgr#start()
	 * @see BufMgr#start()
	 * @see TrxMgr#start()
	 * @throws Exception Thrown if there was a problem starting the Server instance.
	 */
	public void start() throws Exception {
		getLogManager().start();
		getBufferManager().start();
		getTransactionManager().start();
		log.info(LOG_CLASS_NAME, "start", "SIMPLEDBM-LOG: RSS Server started.");
	}
	
	/**
	 * Shuts down the RSS Server instance. This results in following actions.
	 * <ol>
	 * <li>The Transaction Manager is shutdown.</li>
	 * <li>The Buffer Manager is shutdown.</li>
	 * <li>The Log instance is shutdown.</li>
	 * <li>The Storage Manager instance is shutdown.</li>
	 * </ol>
	 * 
	 * @see TrxMgr#shutdown()
	 * @see BufMgr#shutdown()
	 * @see LogMgr#shutdown()
	 * @see StorageManager#shutdown()
	 */
	public void shutdown() {
		getTransactionManager().shutdown();
		getBufferManager().shutdown();
		getLogManager().shutdown();
		getStorageManager().shutdown();
		log.info(LOG_CLASS_NAME, "shutdown", "SIMPLEDBM-LOG: RSS Server shutdown completed.");
	}

	public final IndexManager getIndexManager() {
		return indexManager;
	}

	public final BufMgr getBufferManager() {
		return bufferManager;
	}

	public final LatchFactory getLatchFactory() {
		return latchFactory;
	}

	public final LockMgr getLockManager() {
		return lockManager;
	}

	public final LoggableFactory getLoggableFactory() {
		return loggableFactory;
	}

	public final LogMgr getLogManager() {
		return logManager;
	}

	public final TransactionalModuleRegistry getModuleRegistry() {
		return moduleRegistry;
	}

	public final ObjectFactory getObjectRegistry() {
		return objectRegistry;
	}

	public final PageFactory getPageFactory() {
		return pageFactory;
	}

	public final SpaceMgr getSpaceManager() {
		return spaceManager;
	}

	public final SlottedPageMgr getSlottedPageManager() {
		return slottedPageManager;
	}

	public final StorageContainerFactory getStorageFactory() {
		return storageFactory;
	}

	public final StorageManager getStorageManager() {
		return storageManager;
	}

	public final TrxMgr getTransactionManager() {
		return transactionManager;
	}

	public final TupleManager getTupleManager() {
		return tupleManager;
	}
}
