/***
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *    Project: www.simpledbm.org
 *    Author : Dibyendu Majumdar
 *    Email  : dibyendu@mazumdar.demon.co.uk
 */
package org.simpledbm.rss.bt.impl;

import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.Properties;

import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import org.simpledbm.common.latch.LatchFactory;
import org.simpledbm.common.latch.impl.LatchFactoryImpl;
import org.simpledbm.common.registry.ObjectFactory;
import org.simpledbm.common.registry.impl.ObjectFactoryImpl;
import org.simpledbm.common.util.ByteString;
import org.simpledbm.rss.bm.BufMgr;
import org.simpledbm.rss.bm.BufferAccessBlock;
import org.simpledbm.rss.bm.impl.BufMgrImpl;
import org.simpledbm.rss.bt.IndexKey;
import org.simpledbm.rss.bt.IndexKeyFactory;
import org.simpledbm.rss.bt.IndexScan;
import org.simpledbm.rss.bt.Location;
import org.simpledbm.rss.bt.LocationFactory;
import org.simpledbm.rss.bt.impl.BTreeMgrImpl.BTreeCursor;
import org.simpledbm.rss.bt.impl.BTreeMgrImpl.BTreeImpl;
import org.simpledbm.rss.bt.impl.BTreeMgrImpl.BTreeNode;
import org.simpledbm.rss.bt.impl.BTreeMgrImpl.IndexItem;
import org.simpledbm.rss.io.StorageContainer;
import org.simpledbm.rss.io.StorageContainerFactory;
import org.simpledbm.rss.io.StorageManager;
import org.simpledbm.rss.io.impl.FileStorageContainerFactory;
import org.simpledbm.rss.io.impl.StorageManagerImpl;
import org.simpledbm.rss.locking.LockDuration;
import org.simpledbm.rss.locking.LockMgr;
import org.simpledbm.rss.locking.LockMgrFactory;
import org.simpledbm.rss.locking.LockMode;
import org.simpledbm.rss.locking.impl.LockMgrFactoryImpl;
import org.simpledbm.rss.log.LogMgr;
import org.simpledbm.rss.log.impl.LogFactoryImpl;
import org.simpledbm.rss.pm.Page;
import org.simpledbm.rss.pm.PageFactory;
import org.simpledbm.rss.pm.PageId;
import org.simpledbm.rss.pm.impl.PageFactoryImpl;
import org.simpledbm.rss.sm.SpaceMgr;
import org.simpledbm.rss.sm.impl.SpaceMgrImpl;
import org.simpledbm.rss.sp.SlottedPage;
import org.simpledbm.rss.sp.SlottedPageMgr;
import org.simpledbm.rss.sp.impl.SlottedPageMgrImpl;
import org.simpledbm.rss.tm.LoggableFactory;
import org.simpledbm.rss.tm.Transaction;
import org.simpledbm.rss.tm.TransactionalModuleRegistry;
import org.simpledbm.rss.tm.TrxMgr;
import org.simpledbm.rss.tm.impl.LoggableFactoryImpl;
import org.simpledbm.rss.tm.impl.TransactionalModuleRegistryImpl;
import org.simpledbm.rss.tm.impl.TrxMgrImpl;

public class BTreeTests extends TestCase {

	static final short TYPE_STRINGKEYFACTORY = 25000;
	static final short TYPE_ROWLOCATIONFACTORY = 25001;
	
    boolean doCrashTesting = false;

    public static void main(String[] args) {
		junit.awtui.TestRunner.run(BTreeTests.class);
	}

    public BTreeTests() {
        super();
    }

    public BTreeTests(String arg0) {
        super(arg0);
    }    
    
    public BTreeTests(String arg0, boolean crashTesting) {
        super(arg0);
        doCrashTesting = crashTesting;
    }    
    
	/**
	 * A simple string key.
	 */
	public static class StringKey implements IndexKey {

		static final String MAX_KEY = "<INFINITY>";
		
		ByteString string = new ByteString();
		
		public void setString(String s) {
			parseString(s);
		}
	
		public void setBytes(byte[] bytes) {
			string = new ByteString(bytes);
		}
		
		@Override
		public String toString() {
			if (isMaxKey()) {
				return MAX_KEY;
			}
			return string.toString().trim();
		}
		
		public void parseString(String string) {
			if (MAX_KEY.equals(string)) {
				this.string = new ByteString(new byte[1]);
			}
			else {
				byte data[] = new byte[1024];
				Arrays.fill(data, (byte)' ');
				byte[] srcdata = string.getBytes();
				System.arraycopy(srcdata, 0, data, 0, srcdata.length);
				this.string = new ByteString(data);
			}
		}

		public int getStoredLength() {
			return string.getStoredLength();
		}

		public void retrieve(ByteBuffer bb) {
			string = new ByteString();
			string.retrieve(bb);
		}

		public void store(ByteBuffer bb) {
			string.store(bb);
		}
		
		public boolean isMaxKey() {
			return string.length() == 1 && string.get(0) == 0;
		}
		
		public int compareTo(IndexKey o) {
			StringKey sk = (StringKey) o;
			if (isMaxKey() || sk.isMaxKey()) {
				if (isMaxKey() && sk.isMaxKey()) {
					return 0;
				}
				else if (isMaxKey()) {
					return 1;
				}
				else {
					return -1;
				}
			}
			return string.compareTo(sk.string);
		}

        @Override
        public boolean equals(Object arg0) {
            return compareTo((IndexKey)arg0) == 0;
        }	
	}
	
	public static class StringKeyFactory implements IndexKeyFactory {

		public IndexKey newIndexKey(int id) {
			return new StringKey();
		}

		public IndexKey maxIndexKey(int id) {
			StringKey s = new StringKey();
			s.setBytes(new byte[1]);
			return s;
		}
	}
	
	/**
	 * A simple location.
	 */
	public static class RowLocation implements Location {

		int loc;

		public void parseString(String string) {
			loc = Integer.parseInt(string);
		}

		public void retrieve(ByteBuffer bb) {
			loc = bb.getInt();
		}

		public void store(ByteBuffer bb) {
			bb.putInt(loc);
		}

		public int getStoredLength() {
			return Integer.SIZE/Byte.SIZE;
		}

		public int compareTo(Location o) {
			RowLocation rl = (RowLocation) o;
			return loc - rl.loc;
		}

		@Override
		public boolean equals(Object o) {
			return compareTo((Location)o) == 0;
		}
		
		@Override
		public int hashCode() {
			return loc;
		}

		@Override
		public String toString() {
			return "RowLocation(" + loc + ")";
		}
	}
	
	public static class RowLocationFactory implements LocationFactory {

		public Location newLocation() {
			return new RowLocation();
		}
		
	}

	private IndexItem generateKey(BTreeImpl btree, String s, int location, int childpage, boolean isLeaf) {
		StringKey key = (StringKey) btree.getNewIndexKey();
		key.setString(s);
		RowLocation loc = (RowLocation) btree.getNewLocation();
		loc.loc = location;
		IndexItem item = new IndexItem(key, loc, childpage, isLeaf, btree.isUnique());
		return item;
	}

	private Properties getLogProperties() {
		Properties properties = new Properties();
		properties.setProperty("log.ctl.1", "ctl.a");
		properties.setProperty("log.ctl.2", "ctl.b");
		properties.setProperty("log.groups.1.path", ".");
		properties.setProperty("log.archive.path", ".");
		properties.setProperty("log.group.files", "3");
		properties.setProperty("log.file.size", "16384");
		properties.setProperty("log.buffer.size", "16384");
		properties.setProperty("log.buffer.limit", "4");
		properties.setProperty("log.flush.interval", "5");
		return properties;
	}
	
	/**
	 * Initialize the test harness. New log is created, and the
	 * test container initialized. The container is allocated an extent of
	 * 64 pages which ought to be large enough for all the test cases.
	 */
	void doInitContainer() throws Exception {

		final TestDb db = new TestDb(getLogProperties(), true);
		
		try {
	    	StorageContainer sc = db.storageFactory.create("dual");
	    	db.storageManager.register(0, sc);
	    	Page page = db.pageFactory.getInstance(db.pageFactory.getRawPageType(), new PageId(0, 0));
	    	db.pageFactory.store(page);

	    	db.trxmgr.start();
	    	Transaction trx = db.trxmgr.begin();
			db.spacemgr.createContainer(trx, "testctr.dat", 1, 1, 64, db.spmgr.getPageType());
			trx.commit();
		}
		finally {
			db.shutdown();
		}		
	}
	
	/**
	 * Initialize data pages by loading data from specified XML resource.
	 */
	public void doLoadXml(boolean testUnique, String dataFile) throws Exception {

		final TestDb db = new TestDb(getLogProperties(), false);
		
		try {
	    	StorageContainer sc = db.storageFactory.open("dual");
	    	db.storageManager.register(0, sc);
	    	sc = db.storageFactory.open("testctr.dat");
	    	db.storageManager.register(1, sc);
			
			db.trxmgr.start();

			BTreeMgrImpl.XMLLoader loader = new BTreeMgrImpl.XMLLoader(db.btreeMgr);
			Transaction trx = db.trxmgr.begin();
			try {
				loader.parseResource(dataFile, trx);
			}
			finally {
				// Doesn't matter whether we commit or abort as the changes are
				// logged as redo-only.
				trx.abort();
			}
		}
		finally {
			db.shutdown();
		}		
	}

	/*
	 * Splits page 2 into 2 and 3.
	 */
	void doPageSplit(boolean testLeaf, boolean testUnique) throws Exception {
		/* Create the write ahead log */
		final TestDb db = new TestDb(getLogProperties(), false);
		
		try {
	    	StorageContainer sc = db.storageFactory.open("dual");
	    	db.storageManager.register(0, sc);
	    	sc = db.storageFactory.open("testctr.dat");
	    	db.storageManager.register(1, sc);
			
			final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testUnique);
			db.trxmgr.start();
			
			int pageNumber = 2;
			Transaction trx;
			
			BTreeCursor bcursor = new BTreeCursor();
			bcursor.setQ(db.bufmgr.fixForUpdate(new PageId(1, pageNumber), 0));
			try {
		    	trx = db.trxmgr.begin();
		    	boolean okay = false;
		    	try {
			    	System.err.println("--> SPLITTING PAGE");
					bcursor.searchKey = generateKey(btree, "da", 5, -1, testLeaf);
		    		btree.doSplit(trx, bcursor);
		    	}
		    	finally {
		    		if (okay)
		    			trx.commit();
		    		else {
		    			trx.abort();
		    		}
		    	}
			}
			finally {
				bcursor.unfixQ();
			}
		}
		finally {
			db.shutdown();
		}		
	}
	
	/*
	 * Merges previously split pages 2 and 3.
	 */
	void doRestartAndMerge(boolean testLeaf, boolean testUnique, int l, int r) throws Exception {
		final TestDb db = new TestDb(getLogProperties(), false);

		try {
	    	StorageContainer sc = db.storageFactory.open("dual");
	    	db.storageManager.register(0, sc);
	    	sc = db.storageFactory.open("testctr.dat");
	    	db.storageManager.register(1, sc);

			final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testUnique);
			db.trxmgr.start();

	    	System.err.println("--> BEFORE MERGE");
	    	BufferAccessBlock bab = db.bufmgr.fixShared(new PageId(1, l), 0);
	    	try {
	    		BTreeNode node = btree.getBTreeNode();
	    		node.wrap((SlottedPage) bab.getPage());
	    		node.dump();
	    	}
	    	finally {
	    		bab.unfix();
	    	}
	    	bab = db.bufmgr.fixShared(new PageId(1, r), 0);
	    	try {
	    		BTreeNode node = btree.getBTreeNode();
	    		node.wrap((SlottedPage) bab.getPage());
	    		node.dump();
	    	}
	    	finally {
	    		bab.unfix();
	    	}	    	

	    	BTreeCursor bcursor = new BTreeCursor();
			bcursor.setQ(db.bufmgr.fixForUpdate(new PageId(1, l), 0));
			bcursor.setR(db.bufmgr.fixForUpdate(new PageId(1, r), 0));
			try {
		    	Transaction trx = db.trxmgr.begin();
		    	boolean okay = false;
		    	try {
			    	System.err.println("--> MERGING PAGE");
		    		btree.doMerge(trx, bcursor);
		    	}
		    	finally {
		    		if (okay)
		    			trx.commit();
		    		else {
		    			trx.abort();
		    		}
		    	}
			}
			finally {
				bcursor.unfixQ();
			}
		}
		finally {
			db.shutdown();
		}		
	}

	/*
	 * Merges previously split pages 2 and 3.
	 */
	void doRestart() throws Exception {
		final TestDb db = new TestDb(getLogProperties(), false);
		try {
	    	StorageContainer sc = db.storageFactory.open("dual");
	    	db.storageManager.register(0, sc);
	    	sc = db.storageFactory.open("testctr.dat");
	    	db.storageManager.register(1, sc);

			System.err.println("STARTING RESTART PROCESSING");
			db.trxmgr.start();
	    	System.err.println("RESTART PROCESSING COMPLETED");
		}
		finally {
			db.shutdown();
		}		
	}
	
	
	void doRestartAndLink(boolean testLeaf, boolean testUnique) throws Exception {
		final TestDb db = new TestDb(getLogProperties(), false);
		try {
	    	StorageContainer sc = db.storageFactory.open("dual");
	    	db.storageManager.register(0, sc);
	    	sc = db.storageFactory.open("testctr.dat");
	    	db.storageManager.register(1, sc);

			final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testUnique);
	    	db.trxmgr.start();

	    	System.err.println("--> BEFORE LINKING CHILD TO PARENT");
	    	BufferAccessBlock bab = db.bufmgr.fixShared(new PageId(1, 2), 0);
	    	try {
	    		BTreeNode node = btree.getBTreeNode();
	    		node.wrap((SlottedPage) bab.getPage());
	    		node.dump();
	    	}
	    	finally {
	    		bab.unfix();
	    	}
	    	bab = db.bufmgr.fixShared(new PageId(1, 3), 0);
	    	try {
	    		BTreeNode node = btree.getBTreeNode();
	    		node.wrap((SlottedPage) bab.getPage());
	    		node.dump();
	    	}
	    	finally {
	    		bab.unfix();
	    	}	    	

	    	// prepareParentPage(pageFactory, bufmgr, btree, 1, 4, testUnique, 2);
	    	
	    	BTreeCursor bcursor = new BTreeCursor();
			bcursor.setQ(db.bufmgr.fixForUpdate(new PageId(1, 2), 0));
			bcursor.setR(db.bufmgr.fixForUpdate(new PageId(1, 3), 0));
			bcursor.setP(db.bufmgr.fixForUpdate(new PageId(1, 4), 0));
			try {
		    	Transaction trx = db.trxmgr.begin();
		    	boolean okay = false;
		    	try {
			    	System.err.println("--> LINKING CHILD TO PARENT");
		    		btree.doLink(trx, bcursor);
		    	}
		    	finally {
		    		if (okay)
		    			trx.commit();
		    		else {
		    			trx.abort();
		    		}
		    	}
			}
			finally {
				bcursor.unfixQ();
				bcursor.unfixR();
				bcursor.unfixP();
			}

			bcursor.setQ(db.bufmgr.fixForUpdate(new PageId(1, 2), 0));
			bcursor.setR(db.bufmgr.fixForUpdate(new PageId(1, 3), 0));
			bcursor.setP(db.bufmgr.fixForUpdate(new PageId(1, 4), 0));
			try {
		    	Transaction trx = db.trxmgr.begin();
		    	boolean okay = false;
		    	try {
			    	System.err.println("--> DE-LINKING CHILD FROM PARENT");
		    		btree.doUnlink(trx, bcursor);
		    	}
		    	finally {
		    		if (okay)
		    			trx.commit();
		    		else {
		    			trx.abort();
		    		}
		    	}
			}
			finally {
				bcursor.unfixQ();
				bcursor.unfixR();
			}
		
		}
		finally {
			db.shutdown();
		}		
	}

	void doRestartAndUnlink(boolean testLeaf, boolean testUnique, int p, int q, int r) throws Exception {
		final TestDb db = new TestDb(getLogProperties(), false);
		try {
	    	StorageContainer sc = db.storageFactory.open("dual");
	    	db.storageManager.register(0, sc);
	    	sc = db.storageFactory.open("testctr.dat");
	    	db.storageManager.register(1, sc);

			final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testUnique);
	    	db.trxmgr.start();

	    	BTreeCursor bcursor = new BTreeCursor();
			bcursor.setQ(db.bufmgr.fixForUpdate(new PageId(1, q), 0));
			bcursor.setR(db.bufmgr.fixForUpdate(new PageId(1, r), 0));
			bcursor.setP(db.bufmgr.fixForUpdate(new PageId(1, p), 0));
			try {
		    	Transaction trx = db.trxmgr.begin();
		    	boolean okay = false;
		    	try {
			    	System.err.println("--> DE-LINKING CHILD FROM PARENT");
		    		btree.doUnlink(trx, bcursor);
		    	}
		    	finally {
		    		if (okay)
		    			trx.commit();
		    		else {
		    			trx.abort();
		    		}
		    	}
			}
			finally {
				bcursor.unfixQ();
				bcursor.unfixR();
			}
		}
		finally {
			db.shutdown();
		}		
	}
	
	
	void doRestartAndRedistribute(boolean testLeaf, boolean testUnique) throws Exception {
		final TestDb db = new TestDb(getLogProperties(), false);
		try {
	    	StorageContainer sc = db.storageFactory.open("dual");
	    	db.storageManager.register(0, sc);
	    	sc = db.storageFactory.open("testctr.dat");
	    	db.storageManager.register(1, sc);

			final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testUnique);
	    	db.trxmgr.start();

	    	System.err.println("--> BEFORE REDISTRIBUTING KEYS");
	    	BufferAccessBlock bab = db.bufmgr.fixShared(new PageId(1, 2), 0);
	    	try {
	    		BTreeNode node = btree.getBTreeNode();
	    		node.wrap((SlottedPage) bab.getPage());
	    		node.dump();
	    	}
	    	finally {
	    		bab.unfix();
	    	}
	    	bab = db.bufmgr.fixShared(new PageId(1, 3), 0);
	    	try {
	    		BTreeNode node = btree.getBTreeNode();
	    		node.wrap((SlottedPage) bab.getPage());
	    		node.dump();
	    	}
	    	finally {
	    		bab.unfix();
	    	}	    	

	    	BTreeCursor bcursor = new BTreeCursor();
			bcursor.setQ(db.bufmgr.fixForUpdate(new PageId(1, 2), 0));
			bcursor.setR(db.bufmgr.fixForUpdate(new PageId(1, 3), 0));
			try {
		    	Transaction trx = db.trxmgr.begin();
		    	boolean okay = false;
		    	try {
					bcursor.searchKey = generateKey(btree, "da", 5, -1, testLeaf);
			    	System.err.println("--> REDISTRIBUTING KEYS");
		    		btree.doRedistribute(trx, bcursor);
		    	}
		    	finally {
		    		if (okay)
		    			trx.commit();
		    		else {
		    			trx.abort();
		    		}
		    	}
			}
			finally {
				bcursor.unfixQ();
			}
		}
		finally {
			db.shutdown();
		}		
	}

	void doRestartAndIncreaseTreeHeight(boolean testLeaf, boolean testUnique) throws Exception {
		final TestDb db = new TestDb(getLogProperties(), false);
		try {
	    	StorageContainer sc = db.storageFactory.open("dual");
	    	db.storageManager.register(0, sc);
	    	sc = db.storageFactory.open("testctr.dat");
	    	db.storageManager.register(1, sc);

			final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testUnique);
	    	db.trxmgr.start();

	    	System.err.println("--> BEFORE INCREASING TREE HEIGHT");
	    	BufferAccessBlock bab = db.bufmgr.fixShared(new PageId(1, 2), 0);
	    	try {
	    		BTreeNode node = btree.getBTreeNode();
	    		node.wrap((SlottedPage) bab.getPage());
	    		node.dump();
	    	}
	    	finally {
	    		bab.unfix();
	    	}
	    	bab = db.bufmgr.fixShared(new PageId(1, 3), 0);
	    	try {
	    		BTreeNode node = btree.getBTreeNode();
	    		node.wrap((SlottedPage) bab.getPage());
	    		node.dump();
	    	}
	    	finally {
	    		bab.unfix();
	    	}	    	

	    	BTreeCursor bcursor = new BTreeCursor();
			bcursor.setQ(db.bufmgr.fixForUpdate(new PageId(1, 2), 0));
			bcursor.setR(db.bufmgr.fixForUpdate(new PageId(1, 3), 0));
			try {
		    	Transaction trx = db.trxmgr.begin();
		    	boolean okay = false;
		    	try {
					bcursor.searchKey = generateKey(btree, "da", 5, -1, testLeaf);
			    	System.err.println("--> INCREASING TREE HEIGHT");
		    		btree.doIncreaseTreeHeight(trx, bcursor);
		    	}
		    	finally {
		    		if (okay)
		    			trx.commit();
		    		else {
		    			trx.abort();
		    		}
		    	}
			}
			finally {
				bcursor.unfixQ();
			}
		}
		finally {
			db.shutdown();
		}		
	}

	void doRestartAndDecreaseTreeHeight(boolean testLeaf, boolean testUnique, int p, int q) throws Exception {
		final TestDb db = new TestDb(getLogProperties(), false);
		try {
	    	StorageContainer sc = db.storageFactory.open("dual");
	    	db.storageManager.register(0, sc);
	    	sc = db.storageFactory.open("testctr.dat");
	    	db.storageManager.register(1, sc);

			final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testUnique);
	    	db.trxmgr.start();

	    	System.err.println("--> BEFORE DECREASING TREE HEIGHT");
	    	BufferAccessBlock bab = db.bufmgr.fixShared(new PageId(1, p), 0);
	    	try {
	    		BTreeNode node = btree.getBTreeNode();
	    		node.wrap((SlottedPage) bab.getPage());
	    		node.dump();
	    	}
	    	finally {
	    		bab.unfix();
	    	}
	    	bab = db.bufmgr.fixShared(new PageId(1, q), 0);
	    	try {
	    		BTreeNode node = btree.getBTreeNode();
	    		node.wrap((SlottedPage) bab.getPage());
	    		node.dump();
	    	}
	    	finally {
	    		bab.unfix();
	    	}	    	

	    	BTreeCursor bcursor = new BTreeCursor();
			bcursor.setP(db.bufmgr.fixForUpdate(new PageId(1, p), 0));
			bcursor.setQ(db.bufmgr.fixForUpdate(new PageId(1, q), 0));
			try {
		    	Transaction trx = db.trxmgr.begin();
		    	boolean okay = false;
		    	try {
					bcursor.searchKey = generateKey(btree, "da", 5, -1, testLeaf);
			    	System.err.println("--> DECREASING TREE HEIGHT");
		    		btree.doDecreaseTreeHeight(trx, bcursor);
		    	}
		    	finally {
		    		if (okay)
		    			trx.commit();
		    		else {
		    			trx.abort();
		    		}
		    	}
			}
			finally {
				bcursor.unfixP();
			}
		}
		finally {
			db.shutdown();
		}		
	}

	void doSingleInsert(boolean testUnique, boolean commit, String k, String loc) throws Exception {
		final TestDb db = new TestDb(getLogProperties(), false);
		try {
			StorageContainer sc = db.storageFactory.open("dual");
			db.storageManager.register(0, sc);
			sc = db.storageFactory.open("testctr.dat");
			db.storageManager.register(1, sc);

			final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testUnique);
			db.trxmgr.start();

			IndexKeyFactory keyFactory = (IndexKeyFactory) db.objectFactory.getInstance(TYPE_STRINGKEYFACTORY);
			LocationFactory locationFactory = (LocationFactory) db.objectFactory.getInstance(TYPE_ROWLOCATIONFACTORY);
			IndexKey key = keyFactory.newIndexKey(1);
			key.parseString(k);
			Location location = locationFactory.newLocation();
			location.parseString(loc);
			Transaction trx = db.trxmgr.begin();
			boolean okay = false;
			try {
				System.err.println("--> INSERTING KEY");
				btree.insert(trx, key, location);
				okay = true;
			} finally {
				if (okay && commit)
					trx.commit();
				else {
					trx.abort();
				}
			}
		} finally {
			db.shutdown();
		}		
	}

	void doDoubleInsert(boolean testUnique, boolean commit1, String k1, String loc1, boolean commit2, String k2, String loc2) throws Exception {
		final TestDb db = new TestDb(getLogProperties(), false);
		try {
			StorageContainer sc = db.storageFactory.open("dual");
			db.storageManager.register(0, sc);
			sc = db.storageFactory.open("testctr.dat");
			db.storageManager.register(1, sc);

			final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testUnique);
			db.trxmgr.start();

			IndexKeyFactory keyFactory = (IndexKeyFactory) db.objectFactory.getInstance(TYPE_STRINGKEYFACTORY);
			LocationFactory locationFactory = (LocationFactory) db.objectFactory.getInstance(TYPE_ROWLOCATIONFACTORY);
			
			IndexKey key1 = keyFactory.newIndexKey(1);
			key1.parseString(k1);
			Location location1 = locationFactory.newLocation();
			location1.parseString(loc1);
			Transaction trx1 = db.trxmgr.begin();
			boolean okay1 = false;
			try {
				System.err.println("--> INSERTING KEY");
				btree.insert(trx1, key1, location1);
				okay1 = true;
				
				IndexKey key2 = keyFactory.newIndexKey(1);
				key2.parseString(k2);
				Location location2 = locationFactory.newLocation();
				location2.parseString(loc2);
				Transaction trx2 = db.trxmgr.begin();
				boolean okay2 = false;

				try {
					System.err.println("--> INSERTING KEY");
					btree.insert(trx2, key2, location2);
					okay2 = true;
				}
				finally {
					if (okay2 && commit2)
						trx2.commit();
					else {
						trx2.abort();
					}
				}
			} finally {
				if (okay1 && commit1)
					trx1.commit();
				else {
					trx1.abort();
				}
			}
		} finally {
			db.shutdown();
		}		
	}

	void doSingleDelete(boolean testUnique, boolean commit, String k, String loc) throws Exception {
		final TestDb db = new TestDb(getLogProperties(), false);
		try {
			StorageContainer sc = db.storageFactory.open("dual");
			db.storageManager.register(0, sc);
			sc = db.storageFactory.open("testctr.dat");
			db.storageManager.register(1, sc);

			final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testUnique);
			db.trxmgr.start();

			IndexKeyFactory keyFactory = (IndexKeyFactory) db.objectFactory.getInstance(TYPE_STRINGKEYFACTORY);
			LocationFactory locationFactory = (LocationFactory) db.objectFactory.getInstance(TYPE_ROWLOCATIONFACTORY);
			IndexKey key = keyFactory.newIndexKey(1);
			key.parseString(k);
			Location location = locationFactory.newLocation();
			location.parseString(loc);
			Transaction trx = db.trxmgr.begin();
			boolean okay = false;
			try {
				System.err.println("--> DELETING KEY");
				btree.delete(trx, key, location);
				okay = true;
			} finally {
				if (okay && commit)
					trx.commit();
				else {
					trx.abort();
				}
			}
		} finally {
			db.shutdown();
		}		
	}

	boolean t1Failed = false;
	boolean t2Failed = false;
	
	void doDeleteInsertThreads(final boolean testUnique, final boolean commit, final String k, final String loc) throws Exception {
		final TestDb db = new TestDb(getLogProperties(), false);
    	final boolean testingUniqueIndex = testUnique;
		try {
			StorageContainer sc = db.storageFactory.open("dual");
			db.storageManager.register(0, sc);
			sc = db.storageFactory.open("testctr.dat");
			db.storageManager.register(1, sc);

			db.trxmgr.start();
			
			Thread t1 = new Thread(new Runnable() {
				public void run() {
					try {
						final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testingUniqueIndex);
						IndexKeyFactory keyFactory = (IndexKeyFactory) db.objectFactory.getInstance(TYPE_STRINGKEYFACTORY);
						LocationFactory locationFactory = (LocationFactory) db.objectFactory.getInstance(TYPE_ROWLOCATIONFACTORY);
						IndexKey key = keyFactory.newIndexKey(1);
						key.parseString(k);
						Location location = locationFactory.newLocation();
						location.parseString(loc);
						Transaction trx = db.trxmgr.begin();
						boolean okay = false;
						try {
							System.err.println("--> DELETING KEY");
							btree.delete(trx, key, location);
							Thread.sleep(15000);
							okay = true;
						} finally {
							if (okay && commit)
								trx.commit();
							else {
								trx.abort();
							}
						}
					} catch (Exception e) {
						e.printStackTrace();
						t1Failed = true;
					}
				}
			}, "T1");

			Thread t2 = new Thread(new Runnable() {
				public void run() {
					try {
						final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testingUniqueIndex);
						IndexKeyFactory keyFactory = (IndexKeyFactory) db.objectFactory.getInstance(TYPE_STRINGKEYFACTORY);
						LocationFactory locationFactory = (LocationFactory) db.objectFactory.getInstance(TYPE_ROWLOCATIONFACTORY);
						IndexKey key = keyFactory.newIndexKey(1);
						key.parseString(k);
						Location location = locationFactory.newLocation();
						location.parseString(loc);
						Transaction trx = db.trxmgr.begin();
						boolean okay = false;
						try {
							System.err.println("--> INSERTING KEY");
							btree.insert(trx, key, location);
							okay = true;
						} finally {
							if (okay && commit)
								trx.commit();
							else {
								trx.abort();
							}
						}
					} catch (Exception e) {
						e.printStackTrace();
						t2Failed = true;
					}
				}
			}, "T2");
			
			t1Failed = false;
			t2Failed = false;
			t1.start();
			Thread.sleep(1000);
			t2.start();
			t1.join();
			t2.join();
			assertTrue(!t1.isAlive());
			assertTrue(!t2.isAlive());
			
		} finally {
			db.shutdown();
		}		
	}

    void doScanAndDelete(boolean testUnique, boolean commit, String k, String loc) throws Exception {
		final TestDb db = new TestDb(getLogProperties(), false);
        try {
            StorageContainer sc = db.storageFactory.open("dual");
            db.storageManager.register(0, sc);
            sc = db.storageFactory.open("testctr.dat");
            db.storageManager.register(1, sc);

            final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testUnique);
            db.trxmgr.start();

            IndexKeyFactory keyFactory = (IndexKeyFactory) db.objectFactory.getInstance(TYPE_STRINGKEYFACTORY);
            LocationFactory locationFactory = (LocationFactory) db.objectFactory.getInstance(TYPE_ROWLOCATIONFACTORY);
            IndexKey key = keyFactory.newIndexKey(1);
            key.parseString(k);
            Location location = locationFactory.newLocation();
            location.parseString(loc);
            Transaction trx = db.trxmgr.begin();
            boolean okay = false;
            IndexScan scan = btree.openScan(key, location, LockMode.UPDATE);
            try {
                System.err.println("--> SCANNING TREE");
                while (scan.fetchNext(trx)) {
                	System.err.println("SCAN=" + scan.getCurrentKey() + "," + scan.getCurrentLocation());
                	trx.acquireLock(scan.getCurrentLocation(), LockMode.EXCLUSIVE, LockDuration.MANUAL_DURATION);
                	if (scan.isEof()) {
                		break;
                	}
                	System.err.println("DELETING=" + scan.getCurrentKey() + "," + scan.getCurrentLocation());
                	btree.delete(trx, scan.getCurrentKey(), scan.getCurrentLocation());
                }
            } finally {
                scan.close();
                if (!doCrashTesting) {
					if (okay && commit)
						trx.commit();
					else {
						trx.abort();
					}
				}
            }
        } finally {
			db.shutdown();
        }
    }
	
	
    void doScanTree(boolean testUnique, boolean commit, String k, String loc) throws Exception {
		final TestDb db = new TestDb(getLogProperties(), false);
        try {
            StorageContainer sc = db.storageFactory.open("dual");
            db.storageManager.register(0, sc);
            sc = db.storageFactory.open("testctr.dat");
            db.storageManager.register(1, sc);

            final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testUnique);
            db.trxmgr.start();

            IndexKeyFactory keyFactory = (IndexKeyFactory) db.objectFactory.getInstance(TYPE_STRINGKEYFACTORY);
            LocationFactory locationFactory = (LocationFactory) db.objectFactory.getInstance(TYPE_ROWLOCATIONFACTORY);
            IndexKey key = keyFactory.newIndexKey(1);
            key.parseString(k);
            Location location = locationFactory.newLocation();
            location.parseString(loc);
            Transaction trx = db.trxmgr.begin();
            boolean okay = false;
            IndexScan scan = btree.openScan(key, location, LockMode.SHARED);
            try {
                System.err.println("--> SCANNING TREE");
                while (scan.fetchNext(trx)) {
                	System.err.println("SCAN NEXT=" + scan.getCurrentKey() + "," + scan.getCurrentLocation());
                }
            } finally {
                scan.close();
                if (okay && commit)
                    trx.commit();
                else {
                    trx.abort();
                }
            }
        } finally {
			db.shutdown();
        }
    }

    void doDeleteAndScanThreads(final boolean testUnique, final boolean commit, final String k, final String loc) throws Exception {
		final TestDb db = new TestDb(getLogProperties(), false);
        final boolean testingUniqueIndex = testUnique;
        try {
            StorageContainer sc = db.storageFactory.open("dual");
            db.storageManager.register(0, sc);
            sc = db.storageFactory.open("testctr.dat");
            db.storageManager.register(1, sc);

            db.trxmgr.start();
            
            Thread t1 = new Thread(new Runnable() {
                public void run() {
                    try {
                        final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testingUniqueIndex);
                        IndexKeyFactory keyFactory = (IndexKeyFactory) db.objectFactory.getInstance(TYPE_STRINGKEYFACTORY);
                        LocationFactory locationFactory = (LocationFactory) db.objectFactory.getInstance(TYPE_ROWLOCATIONFACTORY);
                        IndexKey key = keyFactory.newIndexKey(1);
                        key.parseString(k);
                        Location location = locationFactory.newLocation();
                        location.parseString(loc);
                        Transaction trx = db.trxmgr.begin();
                        boolean okay = false;
                        try {
                            System.err.println("--> DELETING KEY [" + k + "," + loc + "]");
                            btree.delete(trx, key, location);
                            Thread.sleep(15000);
                            okay = true;
                        } finally {
                            if (okay && commit) {
                                System.err.println("--> COMMITTING DELETE OF KEY [" + k + "," + loc + "]");
                                trx.commit();
                            }
                            else {
                                System.err.println("--> ABORTING DELETE OF KEY [" + k + "," + loc + "]");
                                trx.abort();
                            }
                        }
                    } catch (Exception e) {
                        e.printStackTrace();
                        t1Failed = true;
                    }
                }
            }, "T1");

            Thread t2 = new Thread(new Runnable() {
                public void run() {
                    try {
                        final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testingUniqueIndex);
                        IndexKeyFactory keyFactory = (IndexKeyFactory) db.objectFactory.getInstance(TYPE_STRINGKEYFACTORY);
                        LocationFactory locationFactory = (LocationFactory) db.objectFactory.getInstance(TYPE_ROWLOCATIONFACTORY);
                        IndexKey key = keyFactory.newIndexKey(1);
                        key.parseString("a1");
                        Location location = locationFactory.newLocation();
                        location.parseString("10");
                        Transaction trx = db.trxmgr.begin();
                        boolean okay = false;
                        IndexScan scan = btree.openScan(key, location, LockMode.UPDATE);
                        try {
                            System.err.println("--> SCANNING TREE");
                            while (scan.fetchNext(trx)) {
                                System.err.println("SCAN=" + scan.getCurrentKey() + "," + scan.getCurrentLocation());
                                if (scan.isEof()) {
                                    break;
                                }
                            }
                        } finally {
                            scan.close();
                            if (okay && commit)
                                trx.commit();
                            else {
                                trx.abort();
                            }
                        }
                    } catch (Exception e) {
                        e.printStackTrace();
                        t2Failed = true;
                    }
                }
            }, "T2");
            
            t1Failed = false;
            t2Failed = false;
            t1.start();
            Thread.sleep(100);
            t2.start();
            t1.join();
            t2.join();
            assertTrue(!t1.isAlive());
            assertTrue(!t2.isAlive());
        } finally {
			db.shutdown();
        }       
    }

    void doScanAndDeleteThreads(final boolean testUnique, final boolean commit, final String k, final String loc) throws Exception {
		final TestDb db = new TestDb(getLogProperties(), false);
        final boolean testingUniqueIndex = testUnique;
        try {
            StorageContainer sc = db.storageFactory.open("dual");
            db.storageManager.register(0, sc);
            sc = db.storageFactory.open("testctr.dat");
            db.storageManager.register(1, sc);

            db.trxmgr.start();
            
            Thread t1 = new Thread(new Runnable() {
                public void run() {
                    try {
                        final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testingUniqueIndex);
                        IndexKeyFactory keyFactory = (IndexKeyFactory) db.objectFactory.getInstance(TYPE_STRINGKEYFACTORY);
                        LocationFactory locationFactory = (LocationFactory) db.objectFactory.getInstance(TYPE_ROWLOCATIONFACTORY);
                        IndexKey key = keyFactory.newIndexKey(1);
                        key.parseString(k);
                        Location location = locationFactory.newLocation();
                        location.parseString(loc);
                        Transaction trx = db.trxmgr.begin();
                        boolean okay = false;
                        try {
                            System.err.println("--> DELETING KEY [" + k + "," + loc + "]");
                            trx.acquireLock(location, LockMode.EXCLUSIVE, LockDuration.MANUAL_DURATION);
                            btree.delete(trx, key, location);
                            okay = true;
                        } finally {
                            if (okay && commit) {
                                System.err.println("--> COMMITTING DELETE OF KEY [" + k + "," + loc + "]");
                                trx.commit();
                            }
                            else {
                                System.err.println("--> ABORTING DELETE OF KEY [" + k + "," + loc + "]");
                                trx.abort();
                            }
                        }
                    } catch (Exception e) {
                        e.printStackTrace();
                        t1Failed = true;
                    }
                }
            }, "T1");

            Thread t2 = new Thread(new Runnable() {
                public void run() {
                    try {
                        final BTreeImpl btree = db.btreeMgr.getBTreeImpl(1, TYPE_STRINGKEYFACTORY, TYPE_ROWLOCATIONFACTORY, testingUniqueIndex);
                        IndexKeyFactory keyFactory = (IndexKeyFactory) db.objectFactory.getInstance(TYPE_STRINGKEYFACTORY);
                        LocationFactory locationFactory = (LocationFactory) db.objectFactory.getInstance(TYPE_ROWLOCATIONFACTORY);
                        IndexKey key = keyFactory.newIndexKey(1);
                        key.parseString("a1");
                        Location location = locationFactory.newLocation();
                        location.parseString("10");
                        IndexKey delkey = keyFactory.newIndexKey(1);
                        delkey.parseString(k);
                        Transaction trx = db.trxmgr.begin();
                        boolean okay = false;
                        IndexScan scan = btree.openScan(key, location, LockMode.UPDATE);
                        try {
                            System.err.println("--> SCANNING TREE");
                            while (scan.fetchNext(trx)) {
                                System.err.println("SCAN=" + scan.getCurrentKey() + "," + scan.getCurrentLocation());
                                System.err.println("Comparing " + scan.getCurrentKey() + " with " + delkey);
                                if (scan.getCurrentKey().equals(delkey)) {
                                    System.err.println("--> SCAN Sleeping for 15 seconds");
                                    Thread.sleep(15000);
                                    System.err.println("--> SCAN Sleep completed");
                                }
                                if (scan.isEof()) {
                                    break;
                                }
                            }
                        } finally {
                            scan.close();
                            if (okay && commit)
                                trx.commit();
                            else {
                                trx.abort();
                            }
                        }
                    } catch (Exception e) {
                        e.printStackTrace();
                        t2Failed = true;
                    }
                }
            }, "T2");
            
            t1Failed = false;
            t2Failed = false;
            t2.start();
            Thread.sleep(1000);
            t1.start();
            t1.join();
            t2.join();
            assertTrue(!t1.isAlive());
            assertTrue(!t2.isAlive());
        } finally {
			db.shutdown();
        }       
    }
    
	public void testPageSplitLeafUnique() throws Exception {
		doInitContainer();
		doLoadXml(true, "org/simpledbm/rss/bt/impl/data1ul.xml");
		doPageSplit(true, true);
	}

//	public void testPageSplitNonLeafUnique() throws Exception {
//		BTreeNode.TESTING_MODE = 2;
//		doInitContainer();
//		doTestXml(true, "org/simpledbm/tests/unit/bt/data1unl.xml");
//		doPageSplit(false, true);
//	}
	
	public void testPageSplitNonLeafUnique2() throws Exception {
		doInitContainer();
		doLoadXml(true, "org/simpledbm/rss/bt/impl/data1unl.xml");
		doPageSplit(false, true);
	}

	public void testRestartAndMerge() throws Exception {
		doRestartAndMerge(false, true, 2, 3);
	}

	public void testPageSplitLeafUnique2() throws Exception {
		doInitContainer();
		doLoadXml(true, "org/simpledbm/rss/bt/impl/data1ul.xml");
		doPageSplit(true, true);
	}
	
	public void testRestartAndLink() throws Exception {
    	System.err.println("--> PREPARING PARENT");
    	doLoadXml(true, "org/simpledbm/rss/bt/impl/data2unl.xml");
		doRestartAndLink(false, true);
	}

	public void testRestartAndLinkAgain() throws Exception {
    	System.err.println("--> PREPARING PARENT");
    	doLoadXml(true, "org/simpledbm/rss/bt/impl/data2unl.xml");
		doRestartAndLink(false, true);
	}

	public void testRestartAndRedistribute() throws Exception {
		doRestartAndRedistribute(false, true);
	}

	public void testRestartAndIncreaseTreeHeight() throws Exception {
		doRestartAndIncreaseTreeHeight(false, true);
	}

	public void testRestartAndUnlink() throws Exception {
		doRestartAndUnlink(false, true, 2, 5, 3);
	}

	public void testRestartAndMergeAgain() throws Exception {
		doRestartAndMerge(false, true, 5, 3);
	}

	public void testRestartAndDecreaseTreeHeight() throws Exception {
		doRestartAndDecreaseTreeHeight(false, true, 2, 5);
	}

	public void testSimpleInsertAbort() throws Exception {
		doInitContainer();
		doLoadXml(false, "org/simpledbm/rss/bt/impl/data3nul.xml");
		doSingleInsert(false, false, "a", "1");
	}
	
	public void testSimpleInsertCommit() throws Exception {
		doInitContainer();
		doLoadXml(false, "org/simpledbm/rss/bt/impl/data3nul.xml");
		doSingleInsert(false, true, "a", "1");
	}
	
	public void testInsertSplitRootAbort() throws Exception {
		doInitContainer();
		// Generate a tree with just the root node which should be full
		doLoadXml(false, "org/simpledbm/rss/bt/impl/data4nul.xml");
		// Following should split the root node
		doSingleInsert(false, true, "da", "8");
		// Following should cause tree height increase
		doSingleInsert(false, false, "b1", "9");
	}
	
	public void testInsertSplitRootCommit() throws Exception {
		doInitContainer();
		// Generate a tree with just the root node which should be full
		doLoadXml(false, "org/simpledbm/rss/bt/impl/data4nul.xml");
		// Following should split the root node
		doSingleInsert(false, true, "da", "8");
		// Following should cause tree height increase
		doSingleInsert(false, true, "b1", "9");
	}

	public void testInsertSplitRootAbortLogical() throws Exception {
		doInitContainer();
		// Generate a tree with just the root node which should be full
		doLoadXml(false, "org/simpledbm/rss/bt/impl/data5nul.xml");
		// This should trigger logical undo as page will be split by
		// second insert
		doDoubleInsert(false, false, "g", "7", true, "da", "8");
	}
	
	public void testInsertUnderflowFig13() throws Exception {
		doInitContainer();
		doLoadXml(false, "org/simpledbm/rss/bt/impl/data6nul.xml");
		doSingleInsert(false, true, "a0", "1");
	}
	
	public void testInsertUnderflowFig14() throws Exception {
		doInitContainer();
		doLoadXml(false, "org/simpledbm/rss/bt/impl/data6nul.xml");
		doSingleInsert(false, true, "c19", "1");
	}

	public void testInsertUnderflowFig5() throws Exception {
		doInitContainer();
		doLoadXml(false, "org/simpledbm/rss/bt/impl/data7nul.xml");
		doSingleInsert(false, true, "k3", "1");
	}
	
	public void testInsertUnderflowFig19() throws Exception {
		doInitContainer();
		doLoadXml(false, "org/simpledbm/rss/bt/impl/data6nul.xml");
		doSingleInsert(false, true, "k3", "1");
	}
	
	public void testInsertUnderflowFig15() throws Exception {
		doInitContainer();
		doLoadXml(false, "org/simpledbm/rss/bt/impl/data6nul.xml");
		doSingleInsert(false, true, "g19", "1");
	}

	// TODO test next key loc across a page
	public void testInsertNextKeyInNextPage() throws Exception {
		doInitContainer();
		doLoadXml(false, "org/simpledbm/rss/bt/impl/data7nul.xml");
		doSingleInsert(false, true, "b4", "1");
	}

	public void testInsertUnderflowFig17() throws Exception {
		doInitContainer();
		doLoadXml(false, "org/simpledbm/rss/bt/impl/data8nul.xml");
		doSingleInsert(false, true, "c0", "1");
	}

	public void testDelete1() throws Exception {
		doInitContainer();
		doLoadXml(false, "org/simpledbm/rss/bt/impl/data6nul.xml");
		doSingleDelete(false, true, "a1", "10");
	}
	
	public void testDeleteInsert1() throws Exception {
		doInitContainer();
		doLoadXml(false, "org/simpledbm/rss/bt/impl/data6nul.xml");
		doDeleteInsertThreads(false, true, "a1", "10");
	}
	
    public void testDeleteInsert2() throws Exception {
		doInitContainer();
		doLoadXml(false, "org/simpledbm/rss/bt/impl/data6nul.xml");
		doDeleteInsertThreads(false, false, "a1", "10");
	}
	
    public void testScan1() throws Exception {
        doInitContainer();
        doLoadXml(false, "org/simpledbm/rss/bt/impl/data6nul.xml");
        doScanTree(false, false, "a1", "10");
    }

	public void testScan2() throws Exception {
        doInitContainer();
        doLoadXml(false, "org/simpledbm/rss/bt/impl/data6nul.xml");
        doScanAndDelete(false, false, "a1", "10");
        doScanTree(false, false, "a1", "10");
        doDeleteAndScanThreads(false, true, "f3", "63");
    }

    public void testScan3() throws Exception {
        doInitContainer();
        doLoadXml(false, "org/simpledbm/rss/bt/impl/data6nul.xml");
        doDeleteAndScanThreads(false, false, "f3", "63");
    }
	
    public void testScan4() throws Exception {
        doInitContainer();
        doLoadXml(false, "org/simpledbm/rss/bt/impl/data6nul.xml");
        doScanAndDeleteThreads(false, false, "f3", "63");
    }

	public void testScanDeleteCrash() throws Exception {
        doInitContainer();
        doLoadXml(false, "org/simpledbm/rss/bt/impl/data6nul.xml");
        doScanAndDelete(false, false, "a1", "10");
    }

	public void testScanAfterCrash() throws Exception {
        doScanTree(false, false, "a1", "10");
    }
	
    public static Test suite() {
        TestSuite suite = new TestSuite();
        suite.addTest(new BTreeTests("testPageSplitLeafUnique"));
        suite.addTest(new BTreeTests("testPageSplitNonLeafUnique2"));
        suite.addTest(new BTreeTests("testRestartAndMerge"));
        suite.addTest(new BTreeTests("testPageSplitLeafUnique2"));
        suite.addTest(new BTreeTests("testRestartAndLink"));
        suite.addTest(new BTreeTests("testRestartAndLinkAgain"));
        suite.addTest(new BTreeTests("testRestartAndRedistribute"));
        suite.addTest(new BTreeTests("testRestartAndIncreaseTreeHeight"));
        suite.addTest(new BTreeTests("testRestartAndUnlink"));
        suite.addTest(new BTreeTests("testRestartAndMergeAgain"));
        suite.addTest(new BTreeTests("testRestartAndDecreaseTreeHeight"));
        suite.addTest(new BTreeTests("testSimpleInsertAbort"));
        suite.addTest(new BTreeTests("testSimpleInsertCommit"));
        suite.addTest(new BTreeTests("testInsertSplitRootAbort"));
        suite.addTest(new BTreeTests("testInsertSplitRootCommit"));
        suite.addTest(new BTreeTests("testInsertSplitRootAbortLogical"));
        suite.addTest(new BTreeTests("testInsertUnderflowFig13"));
        suite.addTest(new BTreeTests("testInsertUnderflowFig14"));
        suite.addTest(new BTreeTests("testInsertUnderflowFig5"));
        suite.addTest(new BTreeTests("testInsertUnderflowFig19"));
        suite.addTest(new BTreeTests("testInsertUnderflowFig15"));
        suite.addTest(new BTreeTests("testInsertNextKeyInNextPage"));
        suite.addTest(new BTreeTests("testInsertUnderflowFig17"));
        suite.addTest(new BTreeTests("testDelete1"));
        suite.addTest(new BTreeTests("testDeleteInsert1"));
        suite.addTest(new BTreeTests("testDeleteInsert2"));
        suite.addTest(new BTreeTests("testScan1"));
        suite.addTest(new BTreeTests("testScan2"));
        suite.addTest(new BTreeTests("testScan3"));
        suite.addTest(new BTreeTests("testScan4"));
        suite.addTest(new BTreeTests("testScanDeleteCrash", true));
        suite.addTest(new BTreeTests("testScanAfterCrash", true));
        return suite;
    }

    public static class TestDb {
        final LogFactoryImpl logFactory;
        final ObjectFactory objectFactory;
        final StorageContainerFactory storageFactory;
        final StorageManager storageManager;
        final LatchFactory latchFactory;
        final PageFactory pageFactory;
        final SlottedPageMgr spmgr;
        final LockMgrFactory lockmgrFactory;
        final LockMgr lockmgr;
        final LogMgr logmgr;
        final BufMgr bufmgr;
        final LoggableFactory loggableFactory;
        final TransactionalModuleRegistry moduleRegistry;
		final TrxMgr trxmgr;
        final SpaceMgr spacemgr;
        final BTreeMgrImpl btreeMgr;

        public TestDb(Properties props, boolean create) throws Exception {

        	logFactory = new LogFactoryImpl();
        	if (create) {
        		logFactory.createLog(props);
        	}
			objectFactory = new ObjectFactoryImpl();
			storageFactory = new FileStorageContainerFactory();
			storageManager = new StorageManagerImpl();
			latchFactory = new LatchFactoryImpl();
			pageFactory = new PageFactoryImpl(objectFactory, storageManager, latchFactory);
			spmgr = new SlottedPageMgrImpl(objectFactory);
			lockmgrFactory = new LockMgrFactoryImpl();
			lockmgr = lockmgrFactory.create(null);
			logmgr = logFactory.getLog(props);
			logmgr.start();
			bufmgr = new BufMgrImpl(logmgr, pageFactory, 5, 11);
			// TODO - move this down
			bufmgr.start();
			loggableFactory = new LoggableFactoryImpl(objectFactory);
			moduleRegistry = new TransactionalModuleRegistryImpl();
			trxmgr = new TrxMgrImpl(logmgr, storageFactory, storageManager, bufmgr, lockmgr, loggableFactory, latchFactory, objectFactory, moduleRegistry);
			spacemgr = new SpaceMgrImpl(objectFactory, pageFactory, logmgr, bufmgr, storageManager, storageFactory, loggableFactory, trxmgr, moduleRegistry);
			btreeMgr = new BTreeMgrImpl(objectFactory, loggableFactory, spacemgr, bufmgr, spmgr, moduleRegistry);
	    	objectFactory.register(TYPE_STRINGKEYFACTORY, StringKeyFactory.class.getName());
			objectFactory.register(TYPE_ROWLOCATIONFACTORY, RowLocationFactory.class.getName());
		}
        
        public void shutdown() {
        	trxmgr.shutdown();
        	bufmgr.shutdown();
			logmgr.shutdown();
			storageManager.shutdown();
        }
    }
    
}
